# -*- coding: utf-8 -*-

import json
import time
import mock
import os
import pytest

from django.conf import settings
from django.core.management import call_command
from django.http.request import HttpRequest, QueryDict
from django.forms.fields import DateField
from django.test.client import RequestFactory, Client
from django.core.urlresolvers import reverse

from mandayejs.mandaye.models import UserCredentials
from mandayejs.mandaye.utils import exec_phantom, get_login_info
from mandayejs.mandaye.forms import FormFactory
from mandayejs.mandaye.views import post_login_do

from utils import create_user, create_credentials, get_uuid, get_user

pytestmark = pytest.mark.django_db


LOGIN_INFO = {
    "address": "https://whatever.com",
    "auth_checker": "tenants/static/js/auth.checker.js",
    "cookies": [],
    "form_submit_element": "input[type=submit], button",
    "locators": [
        {
            "#username": "myusername",
            "#password": "mypassoword"
        }
    ]
}

MOCKED_SITE_LOCATORS = [
    {'id': '#login',
        'label': 'Login',
        'name': 'login',
        'kind': 'string', },
    {'id': '#password',
        'label': 'Password',
        'name': 'password',
        'kind': 'password'},
    {'id': '#birth_date',
        'label': 'Birthdata',
        'name': 'birth_date',
        'kind': 'date'}]


class MockedPopen(mock.Mock):
    stall_for = False

    def communicate(self, data):
        if self.stall_for:
            time.sleep(13)
        return self.expected_output


# ENCRYPTION/DECRYPTION
def test_encryption(credentials):
    decrypted = credentials.decrypt()
    assert decrypted.get('password') == u'jôhn password'


@pytest.fixture(params=['command_ldap', 'command_csv'])
def command(request, command_ldap, command_csv):
    return locals().get(request.param)


@mock.patch('mandayejs.mandaye.management.commands.migrate-users.get_idps')
def test_command_migrate_users(mocked_idps, command):
    mocked_idps.return_value = iter(settings.MELLON_IDENTITY_PROVIDERS)
    call_command(command.name, *command.args, **command.opts)
    if command.opts.get('ldap'):
        credentials = UserCredentials.objects.filter(user__last_name__in=[
            'ldap_user1',
            'ldap_user2',
            'ldap_user3'])

        assert len(credentials) == 3

        for cred in credentials:
            assert cred.to_login_info(decrypt=True)['#password'] == 'password_{}'.format(cred.user.last_name)
    else:
        credentials = UserCredentials.objects.all().exclude(user__last_name__in=[
            'ldap_user1',
            'ldap_user2',
            'ldap_user3'
        ])

        assert len(credentials) == 4

        for cred in credentials:
            assert cred.to_login_info(decrypt=True)['#password'] == cred.to_login_info()['#username']


# MANDAYE API

# GET
def test_api_get(client, url):
    response = client.get(url)
    if client.session.values():
        status_code = 200
    else:
        status_code = 403

    assert response.status_code == status_code

    if status_code == 200:
        assert {'login': '', 'password': ''} == response.data


@pytest.mark.skipif(settings.HOBO is None, reason="hobo is required")
def test_signed_api_get(client_service, url_signed):
    response = client_service.get(url_signed.url)
    if url_signed.orig == 'testserver':
        status_code = 200
    else:
        status_code = 403

    assert response.status_code == status_code

    if status_code == 200:
        assert {'login': '', 'password': ''} == response.data


# POST
@mock.patch('mandayejs.mandaye.api.exec_phantom')
def test_api_post(mock_phantomjs_result, client, url, payload):

    if client.session.values():
        status_code = {'success': 200, 'failure': 401}
    else:
        status_code = {'success': 403, 'failure': 403}

    if payload.get('name_id_content') == '12345':
        response = client.post(url, data=payload, format='json')

        assert response.status_code == status_code['failure']

        if client.session.values():
            kevin = get_user(first_name='kevin')
            assert kevin.username == payload['name_id_content']

            kevin_uuid = get_uuid(name_id=payload['name_id_content'])
            assert kevin_uuid.name_id == '12345'
    else:
        mock_phantomjs_result.return_value = {"result": "ok"}

        response = client.post(url, data=payload, format='json')

        assert response.status_code == status_code['success']

        if client.session.values():
            josh = get_user(username='77777')
            josh_creds = UserCredentials.objects.filter(user=josh)[0]

            assert josh_creds.to_login_info()['#login'] == 'josh'
            assert josh_creds.to_login_info(decrypt=True)['#password'] == 'josh password'


@pytest.mark.skipif(settings.HOBO is None, reason="hobo is required")
@mock.patch('mandayejs.mandaye.api.exec_phantom')
def test_signed_api_post(mock_phantomjs_result, client_service, url_signed, payload):
    if url_signed.orig == 'testserver':
        status_code = {'success': 200, 'failure': 401}
    else:
        status_code = {'success': 403, 'failure': 403}

    if payload.get('name_id_content') == '12345':
        response = client_service.post(url_signed.url, data=payload, format='json')

        assert response.status_code == status_code['failure']

        if url_signed.orig == 'testserver':
            kevin = get_user(first_name='kevin')
            assert kevin.username == payload['name_id_content']

            kevin_uuid = get_uuid(name_id=payload['name_id_content'])
            assert kevin_uuid.name_id == '12345'
    else:
        mock_phantomjs_result.return_value = {"result": "ok"}

        response = client_service.post(url_signed.url, data=payload, format='json')

        assert response.status_code == status_code['success']

        if url_signed.orig == 'testserver':
            josh = get_user(username='77777')
            josh_creds = UserCredentials.objects.filter(user=josh)[0]

            assert josh_creds.to_login_info()['#login'] == 'josh'
            assert josh_creds.to_login_info(decrypt=True)['#password'] == 'josh password'


# DELETE
def test_api_delete(client, url):
    if client.session.values():
        status_code = {'success': 200, 'failure': 404}
    else:
        status_code = {'success': 403, 'failure': 403}

    kevin = get_user(first_name='kevin')
    assert UserCredentials.objects.filter(user=kevin).exists() is False
    response = client.delete(url, data={'name_id_content': '12345'}, format='json')
    assert response.status_code == status_code['failure']

    josh = create_user(username='77777')
    create_credentials(josh, {
        'login': 'josh',
        'password': 'josh password'})

    assert UserCredentials.objects.filter(user=josh).exists() is True
    response = client.delete(url, data={'name_id_content': '77777'}, format='json')
    assert response.status_code == status_code['success']
    if client.session.values():
        assert UserCredentials.objects.filter(user=josh).exists() is False


@pytest.mark.skipif(settings.HOBO is None, reason="hobo is required")
def test_signed_api_delete(client_service, url_signed):
    if url_signed.orig == 'testserver':
        status_code = {'success': 200, 'failure': 404}
    else:
        status_code = {'success': 403, 'failure': 403}

    kevin = get_user(first_name='kevin')
    assert UserCredentials.objects.filter(user=kevin).exists() is False
    response = client_service.delete(url_signed.url, data={'name_id_content': '12345'}, format='json')
    assert response.status_code == status_code['failure']

    josh = create_user(username='77777')
    create_credentials(josh, {
        'login': 'josh',
        'password': 'josh password'
    })

    assert UserCredentials.objects.filter(user=josh).exists() is True
    response = client_service.delete(url_signed.url, data={'name_id_content': '77777'}, format='json')
    assert response.status_code == status_code['success']
    if url_signed.orig == 'testserver':
        assert UserCredentials.objects.filter(user=josh).exists() is False


@mock.patch('mandayejs.mandaye.utils.subprocess.Popen')
@mock.patch('mandayejs.applications.Test.SITE_LOCATORS', MOCKED_SITE_LOCATORS)
def test_phantom_invalid_json(mocked_popen, caplog, user_john):
    expected_output = ('This is not a valid JSON', None)
    mocked_popen.return_value = MockedPopen(expected_output=expected_output)

    UserCredentials.objects.create(user=user_john,
                                   locators={
                                       'login': 'johnny', 'password': 'jumper',
                                       'birth_date': '1995-06-11'})

    client = Client()
    client.login(username='john', password='john')
    response = client.get(reverse('post-login-do'))
    assert 'window.top.location = "/_mandaye/associate/"' in response.content

    for message in response.context['messages']:
        assert message.level_tag == 'error'
        assert message.message == 'invalid response from server'

    for record in caplog.records:
        if record.levelname == 'ERROR':
            assert record.message == 'invalid json: This is not a valid JSON'


@mock.patch('mandayejs.mandaye.utils.subprocess.Popen')
def test_phantom_js_errors(mocked_popen, caplog):
    stdout = {
        "stderr": "ERROR: TypeError: undefined is not a function (evaluating \'$( \"#datepicker\" )",
        "url": "https://whatever.com/someurl",
        "result": "ok"
    }

    expected_output = ('<mandayejs>%s</mandayejs>' % json.dumps(stdout), None)

    mocked_popen.return_value = MockedPopen(expected_output=expected_output)
    result = exec_phantom(LOGIN_INFO)

    for record in caplog.records:
        assert record.levelname == 'WARNING'
        assert record.message == "ERROR: TypeError: undefined is not a function (evaluating \'$( \"#datepicker\" )"

    assert result['result'] == 'ok'
    assert result['url'] == 'https://whatever.com/someurl'


@mock.patch('mandayejs.mandaye.utils.subprocess.Popen')
def test_phantom_js_timeout(mocked_popen, caplog):
    mocked_popen.return_value = MockedPopen(expected_output=json.dumps({'whatever': 'whatever'}), stall_for=True)
    result = exec_phantom(LOGIN_INFO)

    for record in caplog.records:
        assert record.levelname == 'ERROR'
        assert 'https://whatever.com' in record.message
        assert 'tenants/static/js/auth.checker.js' in record.message
        assert 'input[type=submit], button' in record.message

    assert result['result'] == 'timeout'


@mock.patch('mandayejs.applications.Test.SITE_LOCATORS', MOCKED_SITE_LOCATORS)
def test_credentials_json_encoding(user_john):

    request = HttpRequest()
    request.POST = QueryDict('', mutable=True)

    formdata = {
        'login': 'johnny',
        'password': 'jumper',
        'birth_date': '1995-06-11'
    }

    request.POST.update(formdata)

    form = FormFactory(request.POST)
    form.is_valid()

    assert isinstance(form.fields['birth_date'], DateField) is True

    UserCredentials.objects.create(user=user_john, locators=form.cleaned_data)

    cred = UserCredentials.objects.get(user=user_john)

    assert cred.locators['login'] == 'johnny'
    assert cred.locators['birth_date'] == '1995-06-11'


@pytest.fixture(params=[
    {'url1': 'http://mydomain.com/update_password.aspx'},
    {'url2': 'http://mydomain.com/index?path=change_pass'}])
def redirect_url(request):
    return request.param


@mock.patch('mandayejs.mandaye.utils.subprocess.Popen')
@mock.patch('mandayejs.applications.Test.SITE_LOCATORS', MOCKED_SITE_LOCATORS)
def test_password_redirection(mocked_popen, user_john, redirect_url):
    expected_output = {
        "result": "redirect",
        "reason": "password change required",
        "url": redirect_url.get('url1') or redirect_url.get('url2')
    }
    expected_output = '<mandayejs>%s</mandayejs>' % json.dumps(expected_output)
    mocked_popen.return_value = MockedPopen(expected_output=(expected_output, None))

    UserCredentials.objects.create(user=user_john,
                                   locators={
                                       'login': 'johnny', 'password': 'jumper',
                                       'birth_date': '1995-06-11'})

    request = RequestFactory()
    request = request.get(reverse('post-login-do'))
    request.user = user_john
    response = post_login_do(request)
    if 'url1' in redirect_url:
        assert 'window.top.location = "/update_password.aspx"' in response.content
    else:
        assert 'window.top.location = "/index?path=change_pass"' in response.content


@mock.patch('mandayejs.mandaye.utils.subprocess.Popen')
@mock.patch('mandayejs.applications.Test.SITE_LOCATORS', MOCKED_SITE_LOCATORS)
def test_enclosed_response(mocked_popen):
    output = """<mandayejs>{"result": "ok",
    "authentication": "success"} </mandayejs>
    this is just a random error"""

    mocked_popen.return_value = MockedPopen(expected_output=(output, None))
    result = exec_phantom(LOGIN_INFO)
    assert result['result'] == 'ok'

    # with no match
    mocked_popen.return_value = MockedPopen(expected_output=('<mandayejs></mandayejs>', None))
    result = exec_phantom(LOGIN_INFO)
    assert result['result'] == 'json_error'


@mock.patch('mandayejs.mandaye.utils.subprocess.Popen')
@mock.patch('mandayejs.applications.Test.SITE_LOCATORS', MOCKED_SITE_LOCATORS)
def test_post_login_do_with_next_url(mocked_popen, user_john):
    # when sso fails
    expected_output = {
        "result": "redirect",
        "reason": "password change required",
        "url": "http://mydomain.com/update_password.aspx"
    }
    expected_output = '<mandayejs>%s</mandayejs>' % json.dumps(expected_output)
    mocked_popen.return_value = MockedPopen(expected_output=(expected_output, None))

    UserCredentials.objects.create(user=user_john,
                                   locators={
                                       'login': 'johnny', 'password': 'jumper',
                                       'birth_date': '1995-06-11'})

    request = RequestFactory()
    url = '%s?next=http://example.net/' % reverse('post-login-do')
    request = request.get(url)
    request.user = user_john
    response = post_login_do(request)
    assert 'window.top.location = "http://example.net/"' not in response.content

    # when SSO succeeds
    expected_output = {
        "result": "ok",
        "url": "http://mydomain.com/account.aspx"
    }
    expected_output = '<mandayejs>%s</mandayejs>' % json.dumps(expected_output)
    mocked_popen.return_value = MockedPopen(expected_output=(expected_output, None))
    request = RequestFactory()
    url = '%s?next_url=http://example.net/' % reverse('post-login-do')
    request = request.get(url)
    request.user = user_john
    response = post_login_do(request)
    assert 'window.top.location = "http://example.net/"' in response.content


@mock.patch('mandayejs.applications.Test.SITE_LOCATORS', MOCKED_SITE_LOCATORS)
def test_app_settings_overriding(settings, cred_john):
    request = RequestFactory().get('/')
    data = get_login_info(request, cred_john)
    assert data['address'] == 'http://testserver/'
    assert data['auth_checker'] == os.path.join(settings.STATIC_ROOT, 'js/test/auth.checker.js')
    assert data['form_submit_element'] == 'input[type=submit], button'
    # when overriding settings
    settings.SITE_LOGIN_PATH = 'account/login'
    settings.SITE_AUTH_CHECKER = 'js/global.auth.checker.js'
    settings.SITE_FORM_SUBMIT_ELEMENT = 'button'
    data = get_login_info(request, cred_john)
    assert data['address'] == 'http://testserver/account/login'
    assert data['auth_checker'] == os.path.join(settings.STATIC_ROOT, 'js/global.auth.checker.js')
    assert data['form_submit_element'] == 'button'
    settings.SITE_LOGIN_PATH = ''
    settings.SITE_FORM_SUBMIT_ELEMENT = ''
    data = get_login_info(request, cred_john)
    assert data['address'] == 'http://testserver/'
    assert data['form_submit_element'] == ''
